import pygrad.data
import pygrad.exceptions
import pygrad.functions
import pygrad.nn
from pygrad.core import Parameter, Variable, as_variable, no_grad, setup_variable

setup_variable()

__version__ = "0.1.0"
